'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

(function (JsonPropTypes) {
    JsonPropTypes[JsonPropTypes["Any"] = 0] = "Any";
    JsonPropTypes[JsonPropTypes["Boolean"] = 1] = "Boolean";
    JsonPropTypes[JsonPropTypes["Number"] = 2] = "Number";
    JsonPropTypes[JsonPropTypes["String"] = 3] = "String";
})(exports.JsonPropTypes || (exports.JsonPropTypes = {}));

function checkType(value, type) {
    switch (type) {
        case exports.JsonPropTypes.Boolean:
            return typeof value === "boolean";
        case exports.JsonPropTypes.Number:
            return typeof value === "number";
        case exports.JsonPropTypes.String:
            return typeof value === "string";
    }
    return true;
}
function isConvertible(target) {
    if (target && target.prototype) {
        if (target.prototype.toJSON && target.prototype.fromJSON) {
            return true;
        }
        else {
            return isConvertible(target.prototype);
        }
    }
    else {
        return !!(target && target.toJSON && target.fromJSON);
    }
}

class JsonSchemaStorage {
    constructor() {
        this.items = new Map();
    }
    has(target) {
        return this.items.has(target) || !!this.findParentSchema(target);
    }
    get(target) {
        const schema = this.items.get(target) || this.findParentSchema(target);
        if (!schema) {
            throw new Error("Cannot get schema for current target");
        }
        return schema;
    }
    create(target) {
        const schema = { items: {}, target };
        const parentSchema = this.findParentSchema(target);
        if (parentSchema) {
            Object.assign(schema, parentSchema);
            schema.items = Object.assign({}, schema.items, parentSchema.items);
        }
        return schema;
    }
    set(target, schema) {
        this.items.set(target, schema);
        return this;
    }
    findParentSchema(target) {
        const parent = target.__proto__;
        if (parent) {
            const schema = this.items.get(parent);
            return schema || this.findParentSchema(parent);
        }
        return null;
    }
}

const schemaStorage = new JsonSchemaStorage();

class JsonSerializer {
    static serialize(obj, targetSchema, replacer, space) {
        const json = this.toJSON(obj, targetSchema || undefined);
        return JSON.stringify(json, replacer, space);
    }
    static toJSON(obj, targetSchema) {
        let res;
        if (isConvertible(obj)) {
            return obj.toJSON();
        }
        if (Array.isArray(obj)) {
            res = [];
            for (const item of obj) {
                res.push(this.toJSON(item));
            }
        }
        else if (typeof obj === "object") {
            if (targetSchema && !schemaStorage.has(targetSchema)) {
                throw new Error("Cannot get schema for `targetSchema` param");
            }
            targetSchema = (targetSchema || obj.constructor);
            if (schemaStorage.has(targetSchema)) {
                const schema = schemaStorage.get(targetSchema);
                res = {};
                for (const key in schema.items) {
                    const item = schema.items[key];
                    const objItem = obj[key];
                    let value;
                    if ((item.optional && objItem === undefined)
                        || (item.defaultValue !== undefined && objItem === item.defaultValue)) {
                        continue;
                    }
                    if (!item.optional && objItem === undefined) {
                        throw new Error(`Property '${key}' is required in '${obj.constructor.name}' schema`);
                    }
                    if (typeof item.type === "number") {
                        if (item.converter) {
                            if (item.repeated) {
                                value = objItem.map((el) => item.converter.toJSON(el, obj));
                            }
                            else {
                                value = item.converter.toJSON(objItem, obj);
                            }
                        }
                        else {
                            value = objItem;
                        }
                    }
                    else {
                        if (item.repeated) {
                            value = objItem.map((el) => this.toJSON(el));
                        }
                        else {
                            value = this.toJSON(objItem);
                        }
                    }
                    res[item.name || key] = value;
                }
            }
            else {
                res = {};
                for (const key in obj) {
                    res[key] = this.toJSON(obj[key]);
                }
            }
        }
        else {
            res = obj;
        }
        return res;
    }
}

class JsonParser {
    static parse(data, schema) {
        const obj = JSON.parse(data);
        return this.fromJSON(obj, schema);
    }
    static fromJSON(target, targetSchema) {
        const obj = new targetSchema();
        if (isConvertible(obj)) {
            return obj.fromJSON(target);
        }
        const schema = schemaStorage.get(targetSchema);
        for (const key in schema.items) {
            const item = schema.items[key];
            const name = item.name || key;
            const value = target[name];
            if (value === undefined && (item.optional || item.defaultValue !== undefined)) {
                continue;
            }
            if (!item.optional && value === undefined) {
                throw new Error(`Cannot get required property '${name}' for JSON schema '${targetSchema.name}'`);
            }
            if (typeof (item.type) === "number") {
                if (item.converter) {
                    if (item.repeated) {
                        obj[key] = value.map((el) => item.converter.fromJSON(el, obj));
                    }
                    else {
                        obj[key] = item.converter.fromJSON(value, obj);
                    }
                }
                else {
                    if (!item.repeated && !checkType(value, item.type)) {
                        throw new Error(`'${name}' property must be ${exports.JsonPropTypes[item.type]} for schema '${targetSchema.name}'`);
                    }
                    obj[key] = value;
                }
            }
            else {
                if (item.repeated) {
                    obj[key] = value.map((el) => this.fromJSON(el, item.type));
                }
                else {
                    obj[key] = this.fromJSON(value, item.type);
                }
            }
        }
        return obj;
    }
}

const JsonProp = (options = {}) => (target, propertyKey) => {
    const errorMessage = `Cannot set type for ${propertyKey} property of ${target.constructor.name} schema`;
    let schema;
    if (!schemaStorage.has(target.constructor)) {
        schema = schemaStorage.create(target.constructor);
        schemaStorage.set(target.constructor, schema);
    }
    else {
        schema = schemaStorage.get(target.constructor);
        if (schema.target !== target.constructor) {
            schema = schemaStorage.create(target.constructor);
            schemaStorage.set(target.constructor, schema);
        }
    }
    const copyOptions = Object.assign({ type: exports.JsonPropTypes.Any }, options);
    if (typeof copyOptions.type !== "number") {
        if (!schemaStorage.has(copyOptions.type) && !isConvertible(copyOptions.type)) {
            throw new Error(`${errorMessage}. Assigning type doesn't have schema.`);
        }
    }
    schema.items[propertyKey] = copyOptions;
};

exports.JsonSerializer = JsonSerializer;
exports.JsonParser = JsonParser;
exports.JsonProp = JsonProp;
